// Copyright Contributors to the OpenImageIO project.
// SPDX-License-Identifier: Apache-2.0
// https://github.com/AcademySoftwareFoundation/OpenImageIO


// DEPRECATED(2.0)
// Once we were fully C++11 and moved MSVS minimum to >= 2015, the contents
// of this file were no longer needed. What few things were here that
// anybody might still rely on, are in fmath.h.


#pragma once

#include <OpenImageIO/fmath.h>
