#!/usr/bin/env python

# Copyright Contributors to the OpenImageIO project.
# SPDX-License-Identifier: Apache-2.0
# https://github.com/AcademySoftwareFoundation/OpenImageIO

command = rw_command (OIIO_TESTSUITE_IMAGEDIR, "oiio.ico")
