#!/usr/bin/env python

# Copyright Contributors to the OpenImageIO project.
# SPDX-License-Identifier: Apache-2.0
# https://github.com/AcademySoftwareFoundation/OpenImageIO


command += testtex_command ("--maketests 10 \"test{:04}.exr\" --maketest-res 1024 --threadtimes 7 --wedge")

outputs = [ ]
