#!/usr/bin/env python

# Copyright Contributors to the OpenImageIO project.
# SPDX-License-Identifier: Apache-2.0
# https://github.com/AcademySoftwareFoundation/OpenImageIO


# Run the script 
command += pythonbin + " src/test_imagecache.py > out.txt ;"

# compare the outputs
outputs = [ "out.txt" ]

